=head1 NAME

ASCOPE::Term - Aaron's widget for doing things on the command line.

=head1 SUMMARY

 my $term = ASCOPE::Term->new();

 if ($term->ask_yesno("Do something")) {
    my $answer = $term->choose_from_list("Choose me",\@options);
 }

=head1 DESCRIPTION

This is Aaron's widget for doing things on the command line. While
it is true that the world really doesn't need anymore terminal 
thingies, apparently I do.

It is really just a wrapper/helper module for a variety of Term::*
modules with an interface that I<I> like.

For the time being I will probably not send this package to the CPAN.
If people want me to, I will reconsider.

=cut

use strict;
package ASCOPE::Term;

$ASCOPE::Term::VERSION = '0.03';

use Carp;
use Date::Calc qw (Delta_Days check_date);
use Memoize qw (memoize);

use Term::Prompt    qw (termwrap);
use Term::ANSIColor qw (:constants);
$Term::ANSIColor::AUTORESET = 1;

&memoize(
	 "_border",
	 "_chars",
	 "_choicespacer",
	 "_chronological",
	 "_controlspacer",
	 "_dashes",
	 "_default_yesno",
	 "_int",
	 "_is_valid_money_format",
	 "_match_int",
	 "_munge_answer",
	 "_num_dashes",
	 "_valid_day",
	 "_valid_getopts_el",
	 "_valid_month",
	 "_valid_year",
	);

=head1 PACKAGE METHODS

=head2 $pkg = __PACKAGE__->new()

Returns an object. Woot!

=cut

sub new {
  my $pkg  = shift;
  my $self = {};

  bless $self,$pkg;
  return $self;
}

=head1 OBJECT METHODS

=cut

=head2 $pkg->ask($question,$default)

Returns user input for I<$question>.

=cut

sub ask {
  my $self = shift;
  $_[0] || return undef;

  my $answer = $self->_readline($_[0],$_[1]);
  return &_munge_answer("$answer",$_[1]);
}

=head2 $self->ask_yesno($question)

Returns true is answer begins with a "y" or equals "1", 
otherwise returns false.

=cut

sub ask_yesno {
  my $self     = shift;
  my $question = shift;
  my $default  = shift;

  $default = &_default_yesno($default);
  
  my $answer = $self->ask($question,lc $default);
  ($answer =~ /^y/i) ? return 1 : return 0;
}

=head2 $pkg->ask_exists($question, $default)

Requires the user to return *something*. 

=cut

sub ask_exists {
  my $self = shift;
  return $self->ask_match($_[0],$_[1],"[^\\s]+");
}

=head2 $pkg->ask_defined($question,$default)

Requires the user to return something that's been defined
(translation: 0 may not pass a boolean test, but it *is* 
defined). 

=cut

sub ask_defined {
  my $self     = shift;
  my $question = shift;
  my $default  = shift;

  my $answer   = undef;
  
  while (! defined($answer)) {
    $answer = $self->ask($question,$default);

    if (! defined($answer)) {
      print "You must define a value.\n";
    }
  }
  
  return $answer;
}

=head2 $self->ask_match($question,$default,$pattern)

Require the user to return something that matches I<$pattern>.

=cut

sub ask_match {
  my $self     = shift;
  my $question = shift;
  my $default  = shift;
  my $pattern  = shift;

  my $answer   = undef;

  if (! $pattern) {
    carp "You must specify a pattern. Skipping.";
    return undef;
  }
  
  while (! $answer) {
    $answer = $self->ask($question,$default);
    
    if ($answer =~ /$pattern/) {
      $answer = $1;
    } else { 
      print "Not a valid input\n";
      $answer = undef;
    }
  }
  
  return $answer;
}

=head2 $pkg->ask_integer($question,$default)

Prompt the user to return an integer.

Return the int if they do or undef otherwise.

=cut

sub ask_integer {
  my $self     = shift;
  my $question = shift;
  my $default  = shift;
  
  $default = (int($default)) ? $default : undef;
  
  my $answer = $self->ask($question,$default);
  return &_match_int($answer);
}

=head2 $pkg->ask_float($question,$default)

Prompt the user to return an floating point number.

Return the floatif they do or undef otherwise.

=cut

sub ask_float {
  my $self     = shift;
  my $question = shift;
  my $default  = shift;
  
  $default = (&_match_float($default)) ? $1 : undef;
  
  my $answer = $self->ask($question,$default);
  return &_match_float($answer);
}

=head2 $pkg->ask_password($username)

Prompts the user for a password and uses the I<Term::ReadKey> 
package to hide user input.

Returns the password.

=cut

sub ask_password {
  my $self     = shift;
  my $username = shift || "";
  my $pass     = "";
  
  my $prompt = "Please enter password";
  $prompt   .= " for $username" if ($username);

  require Term::ReadKey;

    while (! $pass) {

      $self->_prompt($prompt);

      &Term::ReadKey::ReadMode("noecho");
      $pass = &Term::ReadKey::ReadLine(0);
      chomp $pass;
      &Term::ReadKey::ReadMode("normal");
      print "\n";
    }

    return $pass;
}

=head2 $pkg->ask_date(\@date)

Prompt the use for a valid date.

Returns an array reference containing YYYY,MM,DD

I<The return value for this method may change in 
subsequent releases.>

=cut

sub ask_date {
  my $self = shift;
  my $date = shift;;

  $date ||= [&_today()];

  my $year  = undef;
  my $month = undef;
  my $day   = undef;
  
  while (! $year) {
    $year = $self->ask("Year",($date->[0]));
	
    if (($year) && (! &_valid_year($year))) {
      print "Year failed validation\n";
      $year = undef;
    }
  }

  while (! $month) {
    $month = &ask("Month",$date->[1]);
    
    if (($month) && (! &_valid_month($month))) {
      print "Month failed validation\n";
      $month = undef;
    }
  }

  while (! $day) {
    $day = &ask("Day",$date->[2]);
    
    if (($day) && (! &_valid_day($day))) {
      print "Day failed validation\n";
      $day = undef;
    }
  }
  
#  print "$year $month $day\n";

  if (&_valid_date($year,$month,$day)) {
    return [$year,$month,$day];
  }
  
  if ($self->ask_yesno("Date failed validation. Try again")) {
    return $self->ask_date();
  }
  
  return [];
}

=head2 $pkg->ask_date_span(\@start,\@end)

Prompt the use for two valid dates. Make sure that dates are 
chronological

Returns an array reference containing YYYY-MM-DD, YYYY-MM-DD

I<The return value for this method may change in 
subsequent releases>

=cut

sub ask_date_span {
  my $self  = shift;
  my $start = shift;
  my $end   = shift;
  
  $start = $self->ask_date($start);
  $end   = $self->ask_date($end);
  
  if (&_chronological($start,$end)) {
    return ($start,$end);
  }
  
  if ($self->ask_yesno("Dates were not chronological. Try again")) {
    return $self->ask_date_span();
  }
  
  return ();
}

=head2 $pkg->ask_money_figure($amount,$currency,\@valid_currencies)

Prompt the user for a money amount and a currency.

If no list of valid currencies is provided the user will
br prompted to provide one.

Returns an array reference containing the amount followed
by the currency.

=cut

sub ask_money_figure {
  my $self       = shift;
  my $figure     = shift || "";
  my $currency   = shift || ""; 
  my $currencies = shift;

  my $amount = undef;
  
  while (! $amount) {
    $amount = $self->ask("Amount",$figure);
    
    if (! &_is_valid_money_format($amount)) {
      print "Format failed validation.\n";
      $amount = undef;
    }
  }

  $currency = ($currencies) ? $self->choose_from_list("Currency",$currencies) : $self->ask("Currency",$currency);

  return [$figure,$currency];
}

=head2 $pkg->choose_one($label,\@choices)

Prompt the user to choose one item from a set list.

Lists with eleven, or more, items are presented ten at a time
along with the following options when applicable:

=over

=item *

 >> next

Display next (n) items

=item *

 << prev

Display previous (n) items

=item *

 // done

Return any choices the user may have already made

=item *

 \\ cancel

Return undef

=item *

 || show

Display the list of items the user has chosen

=back

For example :

 213 ->perl -I./lib ./t/005-choose_many.t
 +------------------------------------+
 | 1 - 10 of 11 possible choices      |
 +------------------------------------+
 | * cow                              |
 | * dog                              |
 | * cat                              |
 | * bird                             |
 | * apple                            |
 | * boo                              |
 | * poo                              |
 | * zoo                              |
 | * hump                             |
 | * bump                             |
 +------------------------------------+
 | >> next // done \\ cancel || show  |
 +------------------------------------+
 [choose]

The method uses the I<Term::Complete> package in an effort to 
minimize unnecessary typing.

Returns a string.

=cut

sub choose_one {
  my $self = shift;
  return $self->_choose(@_);
}

=head2 $pkg->choose_many($label,\@choices)

Prompt the user to choose one or more items from a set list.

See docs for $pkg->choose_one() for details.

Returns an array reference.

=cut

sub choose_many {
  my $self = shift;
  return $self->_choose(@_);
}

=head2 $pkg->take_note(\$note)

Prompt the user for a note -- anything that spans multiple 
lines -- or edit an existing one.

The following characters, followed by a new line, are treated
as "special" :

=over 4

=item *

B<.> (dot)

End of note

=item *

B<!.> (bang-dot)

Delete current line

=item *

B<\n> (newline)

Insert new line

=back

When the user passes an existing note, it is split on newlines 
and each is passed to the user for editing.

Returns a string

=cut

sub take_note {
  my $self = shift;
  my $note = shift;

  my $stop = 0;
  my $body = "";
  
  my @previous = (ref($note) eq "SCALAR") ? split("\n",$$note) : ();
  
  while (! $stop) {
    
    my $line   = "";
    my $prompt = "new";

    if (@previous) {
      $line   =  shift @previous;
      $prompt = "old";
    }
    
    my $input = $self->_readline("$prompt ",$line);
    
    if ((! $input) && ($prompt eq "old")) {
      $body = "$line\n";
    }

    # End of note
    elsif ($input =~ /^(\.)$/) {
      chomp $body;
      $stop = 1;
      
      # Delete this line
    } 

    elsif ($input =~ /^(!\.)$/) {
      next;
      
      # Append this line
    } 

    else {
      next if ($body eq $input);
      
      # Check to see if the user has asked 
      # for one, or more, new lines to be inserted
      # in the current one
      
      $input =~ s/\\n/\n/gm;
      $body .= "$input\n" 
    }
  }
  
  return $body;
}

=head2 $self->ask_getopts(\%opts,@order)

Prompt the user for bunch of stuff using Getopt::Long.

I<$opts> is the associative array. 

The key is expected to use the same format that L<Getopt::Long::GetOptions> 
uses for keys.

The value for the key is an array reference whose values are :

=over

=item *

B<result>

Reference. Required.

This is a reference to the variable where the result is stored.

=item *

B<prompt>

String.

If present this string will be printed to STDOUT when prompting the 
user for input. If not defined, the the key name for the current array 
ref will be used.

=item *

B<default>

Code reference.

This can be used to define a default value based on previous user input 
that is not already available at compile time.

The return value of your anonymous sub must be passed as a reference. Currently, 
the function is only smart about handling scalar refs.

If you return an array ref, for instance, hoping that the user will be required to 
choose from that list, you'll be disappointed. It's on the list.

If no default is provided then the default value is assumed to be any value already 
defined in I<result>.

=back

If I<@order> is not defined, then the user will be prompted for each key in I<$opts> 
alphabetically.

Example:

 my %opts = (
	     "username"  => "bob",
	     "email"     => undef,
 	    );


 $term->ask_getopts({
	 	     "username=s" => [ \$opts{'username'}],
		     "email=s" => [
				    \$opts{'email'},
				    "email address",
				    sub { return \"$opts{'username'}\@bob.com"; },
				  ],
                                 "username","email",
                                });

 # would produce:

 asc @ localhost in /home/asc/bin/perl/debug
 122 ->./term-ref
 username?  [bob] asc
 email address? [asc@bob.com] 
 $VAR1 = {
           'email' => 'asc@bob.com',
           'username' => 'asc',
         };

This method still isn't very smart about lists.

=cut

sub ask_getopts {
    my $self  = shift;
    my $opts  = shift;
    my @order = @_;

    my %types = ();

    foreach my $el (keys %$opts) {
      my $valid = &_valid_getopts_el($el);

      if (! $valid) {
	warn "Failed to parse key '$el'. Skipping.";
	next;
      }

      $types{ $valid->{key} }->{ "op" }   = $valid->{op};
      $types{ $valid->{key} }->{ "type" } = $valid->{type};
    }

    if (! @order) { @order = sort keys %types; }

    foreach my $el (@order) {

	next unless ($types{ $el });

	my $operator = $types{ $el }->{ "op" };
	my $type     = $types{ $el }->{ "type" };
	my $optkey   = join($operator,$el,$type);

	my $optargs  = $opts->{ $optkey };

	if (ref($optargs) ne "ARRAY") {
	  warn "Arguments for key $el is not an array ref. Skipping.";
	  next;
	}

	#

	my $result   = $optargs->[0];
	my $deftype  = ref($result);

	if (! $deftype) {
	    warn "'$optkey' is not a reference. Skipping.";
	    next;
	}

	#

	my $question = $optargs->[1] || $el;

	#

	my $default  = $optargs->[2];

	if (($default) && (ref($default))) {
	  if (ref($default) ne "CODE") { 
	    warn "default key is not a CODE reference. Skipping.";
	    next;
	  }

	  $default = $self->$default();
	}

	#

	$default ||= $result;

	#

	if ((! $operator) && (! $type)) {
	    $self->_ask_opt(
		      $question,
		      "ask_yesno",
		      $result,
		      $default,
		      $deftype,
		      );
	    next;
	}

	if (! $operator) {
	    $self->_ask_opt(
		      $question,
		      "ask",
		      $result,
		      $default,
		      $deftype,
		      );
	    next;
	}

	if ($operator eq "!") {
	    $self->_ask_opt(
		      $question,
		      "ask_yesno",
		      $result,
		      $default,
		      $deftype,
		      );
	    next;
	}
	
	if ($operator eq "+") {
	    my $num = undef;
	    while ($self->ask_yesno($question)) {
		$num ++;
	    }
	    
	    &_set_optref($default,$num);
	    next;
	}

	if (($type eq "s") && ($operator eq ":")) {
	  $self->_ask_opt(
		      $question,
		      "ask",
		      $result,
		      $default,
		      $deftype,
		      );
	    next;
	}

	if (($type eq "s") && ($operator eq "=")) {
	  $self->_ask_opt(
		      $question,
		      "ask",
		      $result,
		      $default,
		      $deftype,
		      "exists",
		      );
	    next;
	}

	if (($type eq "i") && ($operator eq ":")) {
	  $self->_ask_opt(
		      $question,
		      "ask_integer",
		      $result,
		      $default,
		      $deftype,
		      );
	    next;
	}

	if (($type eq "i") && ($operator eq "=")) {
	    $self->_ask_opt(
		      $question,
		      "ask_integer",
		      $result,
		      $default,
		      $deftype,
		      "defined",
		      );
	    next;
	}

	if (($type eq "f") && ($operator eq ":")) {
	    $self->_ask_opt(
		      $question,
		      "ask_float",
		      $result,
		      $default,
		      $deftype,
		      );
	    next;
	}

	if (($type eq "f") && ($operator eq "=")) {
	    $self->_ask_opt(
			    $question,
			    "ask_float",
			    $result,
			    $default,
			    $deftype,
			    "defined",
			   );
	    next;
	  }

    }
    
    return $opts;
}

# Private methods

sub _readline {
  my $self = shift;
 
  if (@_) { $self->_prompt(@_); }

  my $answer;

  {
    local($/) = "\n";
    $answer = <STDIN>;
    chomp($answer);
  }

  return $answer;
}

sub _prompt {
  my $self = shift;
  &_print_coloured("BOLD",(($_[1]) ? "$_[0] [$_[1]] " : "$_[0] "));
}

sub _print_coloured {

  if ($ENV{EMACS}) {
    print &termwrap($_[1]);
  }

  # To be finessed...
  elsif ($_[0] eq "BOLD") {
    print BOLD &termwrap($_[1]);
  }

}

sub _choose {
  my $self    = shift;
  my $label   = shift;
  my $actions = shift;

  if (ref($actions) ne "ARRAY") {
    print "Actions must be passed as an array reference.\n";
    return undef;
  }

  require Term::Complete;

  #

  my $caller   = (caller(1))[3];
  $caller      =~ s/.*:://;

  my $stop     = 0;
  my $continue = 1;

  my $choice  = undef;
  my @choices = ();

  my $previous = "<< previous";
  my $next     = ">> next";
  my $finish   = "// done";
  my $exit     = "\\\\ cancel";
  my $show     = "|| show";

  my $count     = scalar @$actions;
  my $increment = 9;
  my $first     = 0;

  #

  while ($continue) {

    while (! $choice) {

      my $last = $first + $increment;
      if ($last > $count) { $last = ($count -1); }

      my $str_choices      = ($first+1)." - ".(($last == $count) ? $last : ($last+1))." of $count possible choices";
      my $chars_strchoices = length($str_choices);
      #

      my @controls = ();
      
      if ($first > $increment) {
	push @controls,$previous;
      }
      
      if ($last+1 < $count) {
	push @controls,$next;
      }
      
      push @controls,$finish,$exit;

      if ($caller eq "choose_many") {
	push @controls, $show;
      }

      my $str_controls      = join(" ",@controls);
      my $chars_strcontrols = length($str_controls);

      #

      $choice = &_choose_menu($label,$str_choices,\@controls,[@{$actions}[$first..$last]]);

      #

      if ($choice eq $previous) {
	$first -= ($increment + 1);
	
	$choice = undef;
	next;
      }
      
      if ($choice eq $next) {
	$first = $last + 1;
	
	$choice = undef;
	next;
      }
      
      if ($choice eq $finish) {
	return ($caller eq "choose_one") ? $choices[0] : \@choices;
      }

      if ($choice eq $exit) {
	return undef;
      }

      #

      if ($caller eq "choose_one") {

	if (! $choice) { return undef; }

	if (! grep /^($choice)$/,@{$actions}[$first..$last]) {
	  print "[Unknown action]\n";
	  $choice = undef;
	  next;
	}

	return $choice;
      }

      elsif ($choice) {

	if ($choice eq $show) {
	  &_display_menu("Items chosen",\@controls,\@choices);

	  $choice = undef;
	  next;
	}

	if (! grep /^($choice)$/,@{$actions}[$first..$last]) {
	  print "[Unknown action]\n";
	  $choice = undef;
	  next;
	}
	
	push @choices,$choice;
	$choice = undef;
      }
    }
  }

  return \@choices;
}

sub _choose_menu {
  my $label       = shift;
  my $str_choices = shift;
  my $controls    = shift;
  my $choices     = shift;

  &_display_menu($str_choices,$controls,$choices);

  my $str_controls = join(" ",@$controls);
  my $spacer       = &_controlspacer(&_chars($str_choices,$str_controls),$choices);

  print
    "| ";

  &_print_coloured("BOLD",$str_controls);

  print 
    "$spacer |\n",
    &_border(&_chars($str_choices,$str_controls),$choices);

  &_print_coloured("BOLD","[$label] ");
  return Term::Complete::Complete(undef,[@$choices,@$controls]);
}

sub _display_menu {
  my $str_choices = shift;
  my $controls    = shift;
  my $choices     = shift;

  my $str_controls      = join(" ",@$controls);

  my $spacer = &_choicespacer(&_chars($str_choices,$str_controls),$choices);
  my $border = &_border(&_chars($str_choices,$str_controls),$choices);
  my $dashes = &_num_dashes(&_chars($str_choices,$str_controls),$choices);

  print
    $border,
    "| ";

  &_print_coloured("BOLD",$str_choices);

  print
    "$spacer |\n",
    $border;

  foreach (@$choices) {
    print "| * $_";
    map { print " "; } (2..($dashes - length($_)));
    print " |\n";
  }

  print $border;
}

# memoized

sub _choicespacer {
  my $chars_strchoices  = shift;
  my $chars_strcontrols = shift;
  my $choices           = shift;

  my $num_dashes = &_num_dashes($chars_strchoices,$chars_strcontrols,$choices);
  my $spacer     = "";

  map { $spacer .= " "; } (0..($num_dashes - $chars_strchoices));
  return $spacer;
}

sub _controlspacer {
  my $chars_strchoices  = shift;
  my $chars_strcontrols = shift;
  my $choices           = shift;

  my $num_dashes = &_num_dashes($chars_strchoices,$chars_strcontrols,$choices);
  my $spacer     = "";

  map { $spacer .= " "; } (0..($num_dashes - $chars_strcontrols));
  return $spacer;
}

# memoized

sub _chars {
  return (length($_[0]),length($_[1]));
}

# memoized

sub _border {
  return "+-".&_dashes(@_)."-+\n";
}

# memoized

sub _dashes {
  my $dashes = undef;
  map { $dashes .= "-"; } (0..&_num_dashes(@_));
  return $dashes;
}

# memoized

sub _num_dashes {
  my $chars_strchoices  = shift;
  my $chars_strcontrols = shift;
  my $choices           = shift;

  my $num_dashes = ($chars_strchoices > $chars_strcontrols) ? $chars_strchoices : $chars_strcontrols;

  map { if (length($_) > $num_dashes) { $num_dashes = length($_); } } @$choices;
  return $num_dashes;
}

#

sub _ask_opt {
    my $self     = shift;
    my $el       = shift;
    my $action   = shift;
    my $result   = shift;
    my $default  = shift;
    my $deftype  = shift;
    my $required = shift;

    my $option   = undef;

    # print "DEBUG $el ($deftype) \n";

    if ($deftype eq "SCALAR") {

        $option = $self->_do_ask_opt($el,$action,$default,$required);
	&_set_optref($result,$option);

	return 1;
    }
    
    if ($deftype =~ /^(ARRAY|HASH)$/) {
	
	my $continue = 1;
	my @input    = ();
	
	while ($continue) {
	    
	    if ($deftype eq "ARRAY") {

		$option = $self->_do_ask_opt($el,$action,$default,$required);
		push(@input, $option);

	    } 
	    
	    else {

		my $key   = undef;
		my $value = undef;

		while (! $key) {
		    $key = $self->ask("Key");

		    if (! $key) {
			print "You must specify a key.\n";
		    }
		}

		$option = $self->_do_ask_opt($key,$action,undef,$required);
		push(@input,$key=>$option);
	    }

	    $continue = $self->ask_yesno("Add another value for $el");
	}
	
	&_set_optref($result,@input);
	return 1;
    }
    
    return 0;
}

sub _do_ask_opt {
    my $self     = shift;
    my $el       = shift;
    my $action   = shift;
    my $default  = shift;
    my $required = shift;

    # print "DEBUG $el ($action) $required \n";

    no strict "refs";
    my $str = undef;

    if ($required eq "exists") {
	while (! $str) {
	    $str = $self->$action($el,&_get_optref($default));
	    
	    if (! $str) {
		print "Not a valid input for $el.\n";
	    }
	}
    } 

    elsif ($required eq "defined") {
	while (! defined($str)) {
	    $str = $self->$action($el,&_get_optref($default));
	    
	    if (! defined($str)) {
		print "Not a valid input for $el.\n";
	    }
	}
    } 

    else {
	$str = $self->$action($el,&_get_optref($default));
    }
    
    return $str;
}

sub _get_optref {
    my $ref  = shift;
    my $type = ref($ref);

    if ($type eq "SCALAR") {
	return $$ref;
    }

    if ($type eq "ARRAY") {
	return @{$ref};
    }
    
    if ($type eq "HASH") {
	return %{$ref};
    }

    carp "Unknown or invalid ref type : '$ref'";
    return 0;
}

sub _set_optref {
    my $ref  = shift;
    my @args = @_;
    my $type = ref($ref);

    # print "DEBUG setting '$type' ".join(",",@args)."\n";

    if ($type eq "SCALAR") {
	$$ref = $args[0];
	return 1;
    }

    if ($type eq "ARRAY") {
	@{$ref} = @args;
	return 1;
    }

    if ($type eq "HASH") {
	%{$ref} = @args;
	return 1;
    }

    carp "Unknown or invalid ref type : '$ref'";
    return 0;
}

# memoized

sub _munge_answer {
    my $answer  = shift;
    my $default = shift;

    # Strip default
    $answer =~ s/^((\s)*\[([^]]+)])\s//;

    if ($answer eq "")        { return $default; }
    if ($answer =~ /^(\s+)$/) { return " ";      }

    $answer =~ s/^(\s)+//;
    $answer =~ s/(\s)+$//;

    $answer = ($answer =~ /^$/) ? $default : $answer;
    return $answer;
}

sub _today {
    my @today = (localtime)[5,4,3];
    return [$today[0]+1900,$today[1]+1,$today[2]];
}

# memoized

sub _chronological {
    my $start = shift;
    my $end   = shift;

    my $test  = undef;

    # START BEFORE IT ENDS
    $test = &Date::Calc::Delta_Days(@$start,@$end);
    return undef if ($test < 0);

    # ENDS BEFORE IT STARTS
    $test = &Date::Calc::Delta_Days(@$end,@$start);
    return undef if ($test > 0);

    return 1;
}

# memoized

sub _valid_year {
    my $n = shift;

    return 0 if (! &_int($n));
    return 0 if ($n < 1);

    return 1;
}

# memoized

sub _valid_month {
    my $n = shift;

    return 0 if (! &_int($n));
    return 0 if ($n < 1);
    return 0 if ($n > 12);

    return 1;
}

# memoized

sub _valid_day {
    my $n = shift;

    return 0 if (! &_int($n));
    return 0 if ($n < 1);
    return 0 if ($n > 31);

    return 1;
}

# memoized

sub _valid_date {
    my $date  = (@_)[0,1,2];
    return &Date::Calc::check_date(@$date);
}

# memoized

sub _int {
  ($_[0] =~ /^(\d+)$/) ? return 1 : return 0;
}

# memoized

sub _match_int {
  ($_[0] =~ /^((\+|-)*(\d+))$/) ? return $1 : return undef;
}

# memoized

sub _match_float {
  ($_[0] =~ /^((\+|-)*(\d+)(\.)*(\d+)*)$/) ? return $1 : return undef;
}

# memoized

sub _is_valid_money_format {
  ($_[0] =~ /^(\d+)*\.(\d\d)*$/) ? return 1 : return 0;
}

# memoized

sub _default_yesno {
  if ($_[0] eq "0") {
    return "n";
  }
  
  elsif ($_[0] eq "1") {
    return "y";
  }
  
  elsif ($_[0] =~ /^(y|n)$/i) {
    return $1;
  }
  
  else {
    return "y";
  }

}

# memoized

sub _valid_getopts_el {
   $_[0] =~ /^(([^!\+:=]+)((!|\+)|((:|=)(f|i|s)))*)$/;
   ($1) ? return {key=>$2,op=>$6,type=>$7} : return 0;
}

=head1 VERSION 

0.03

=head1 DATE

December 30, 2002

=head1 AUTHOR 

Aaron Straup Cope

=head1 REQUIREMENTS

L<Term::Prompt>

L<Term::ANSIColor>

L<Term::ReadKey>

L<Term::Complete>

L<Date::Calc>

L<Memoize>

=head1 BUGS

This release works much better in an Emacs shell than earlier versions
however there are still bugs. It basically doesn't work at all in
I<eshell> mode.

=head1 LICENSE

Copyright (c) 2001-2002, Aaron Straup Cope. All Rights Reserved.

This is free software, you may use it and distribute it under the same terms as Perl itself.

=cut

return 1;
